""" ARDI Historian Driver Base Class

The 'Base' class encapsulates all of the essential startup and control behaviours of an ARDI historical driver.

Each 'Base' contains multiple 'Core' objects, which in turn contain a single source instance.

"""
import platform

import sys
import logging
import threading
import historianipc
import histdrivercore
import argparse
import pycurl
import xmltodict
from StringIO import StringIO
from twisted.web import server,http,resource

def corethread(core, servername, assetid):
    core.Launch(servername, assetid)

def webservice(core):
    """ Starts the webservice used to communicate with and manage this driver

    \param core: The base driver that is controlled by this web service"""
    core.logger.info('Starting IPC Service')    
    
    ipc = historianipc.DriverIPC()
    ipc.core = core
    ipc.port = core.hostport
    if ipc.port == 0:
        ipc.port = 8081
        
    try:
        ipc.start()
    except KeyboardInterrupt:
        pass
        
class historian:
    """ The base class for a historian driver.

    This class is responsible for all stages of driver life-cycle.

    It also communicates directly with ARDI to get details on what data sources it should serve via HTTP web requests """
    
    def __init__(self):
        """Initialises the historian"""
        ##The IPC port for this driver
        self.hostport = 10000        
        pass
    
    def usage(self):
        """Prints usage information if startup fails"""
        print "Historian Startup: <drivername> <asset1>,<asset2>...<assetn>"
        return
    
    def initlogger(self):
        """Initialises the historian logger"""
        logging.basicConfig(level=logging.INFO)
        ##The logger used to record information about the driver process as a whole
        self.logger = logging.getLogger(__name__)
        
        fname = "/var/log/ardi/driver-" + str(self.hostport) + ".log"
        if (platform.system() == 'Windows'):
            fname = "c:\\windows\\temp\\ardi-driver-" + str(self.hostport) + ".log"

        loghandler = logging.FileHandler(fname)
        loghandler.setLevel(logging.DEBUG)

        # create a logging format
        
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        loghandler.setFormatter(formatter)
        
        # add the handlers to the logger
        
        self.logger.addHandler(loghandler)

        self.logger.info('Starting Historian Process')

    def start(self,driverfactory):
        """Start the Driver

        This function is called once all initialisation is complete. It launches sources and the IPC service.

        \param driverfactory: The factory class that produces driver instances"""
        self.initlogger()

        ##A class that generates historian drivers on request
        self.factory = driverfactory

        arglist = sys.argv
        if len(arglist) >= 2:
            if arglist[1] == 'start':                
                arglist = arglist[2:]
            else:
                arglist = arglist[1:]
  
        parser = argparse.ArgumentParser(description='ARDI Driver')
        parser.add_argument('driverid',metavar='PortNo',type=int, help='Unique ID/port for the driver', nargs='?')
        parser.add_argument('server',metavar='Server', help='Server Name', nargs='?', default='localhost')
        parser.add_argument('assetid',metavar='AssetNo',type=int, help='Asset IDs to act as a driver for', nargs='?')
        parser.add_argument('--multiple',dest='multi',action='store_const',const=True,default=False,help='Allows this driver to connect to multiple devices')        
        parser.add_argument('--script','-r',dest='script',default='')
        parser.add_argument('--shell','-l',dest='shell',default='')
        parser.add_argument('--config','-c',dest='configfile',default='')        
        parser.add_argument('--debuglevel',dest='debuglevel',default=2)
        args = parser.parse_args(arglist)
        
        self.logger.debug(args)        
        
        servername = args.server
        self.hostport = int(args.driverid)
        
        self.logger.info('Driver ' + str(self.hostport) + " feeds ARDI server " + servername)

        #self.script = args.script
        #self.shell = args.shell        

        ##A list of the currently active ARDI sources
        self.drivers = {}

        ##A quick lookup dictionary to map addresses to ARDI names
        self.assetids = []
        if args.assetid != None:
            for assetid in args.assetid:
                self.assetids.append(assetid)
        else:
            response = StringIO()
        
            fullurl = "http://" + servername + "/api/driverdata?port=" + str(self.hostport) + "&mode=1"
        
            c = pycurl.Curl()
            c.setopt(c.URL,fullurl)
            c.setopt(c.WRITEFUNCTION,response.write)
            c.setopt(c.TIMEOUT,5)
            c.setopt(c.CONNECTTIMEOUT,5)
            c.perform()
            c.close()

            print response.getvalue()

            self.parseconfig(response.getvalue())
            
            print "Load Default Configuration From Server"

        for assetid in self.assetids:
            self.logger.info("Creating historian driver instance for " + str(assetid))
            instance = driverfactory.createinstance()
            core = histdrivercore.historiancore(instance)
            core.base = self
            core.source = args.configfile    
            self.drivers[str(assetid)] = core
            core.Launch(servername,assetid)

        #print self.drivers
        webservice(self)

    def parseconfig(self,xml):
        """Imports the configuration XML file

        This function loads the list of data points within the XML text provided. The function is usually called from LoadConfig

        \param xml: The configuration data in XML format"""
        result = xmltodict.parse(xml)
                        
        #Core Parameters
        try:            
            if '@id' in result['driverdata']['source']:
                self.assetids.append(result['driverdata']['source']['@id'])
            else:
                for itm in result['driverdata']['source']:
                    self.assetids.append(itm['@id'])
        except:
            pass
            
        #print result
                
    def Query(self,assetid, points,start,end,function,grain,first):
        """Performs a query on a single source

        \param assetid: The SOURCE asset id
        \param points: A list of points to query
        \param start: The start date for the query, in YYYY-MM-DD hh:mm:ss format
        \param end: The start date for the query, in YYYY-MM-DD hh:mm:ss format
        \param function: The function to perform. Functions include 'raw', 'avg', 'sum', 'min' and 'max'
        \param grain: The number of seconds between samples, or '0' for natural sampling rate
        \param first: True if this the data is to include the samples immediately preceeding the start time.
        """
        return self.drivers[assetid].Query(points,start,end,function,grain,first)
    
    def Reload(self, assetid):
        """Forces a historian source to reload

        This function is usually called internally by a change to the ARDI data.

        \param assetid: The SOURCE asset id            
        """
        self.drivers[assetid].Reload()

    def Restart(self):
        """Forces the entire driver process to reload

        TODO: Note that this is currently unimplemented for historian drivers!
        
        """
        self.logger.info('Driver Restart Requested via IPC')
        #self.StopDrivers()
        #self.LoadConfig()