import pycurl
import xmltodict
import time
import socket
import transform
from subprocess import Popen
import subprocess

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from StringIO import StringIO

class pointinfo:
    """Data Point Information

    An empty class used for storing information about data points"""
    pass

class historiancore:
    """A connection to a single instance or device

    Unlike live drivers, historian drivers are mostly passive and don't run their own individual threads.

    They wait until they are queried by the system to return data.
    """
    
    def __init__(self,driver):
        """ Initialises the Core"""
        ##The url to return data to, used for remote drivers
        self.url = "localhost"

        ##The source ID
        self.asset = 1

        ##The actual driver instance - the class that performs the query on the data store
        self.driver = driver

        ##A self-reference        
        driver.core = self        
    
    def SetARDIDetails(self, url, assetid):
        """Set Connection Details

        This function sets some of the basic identity information about the core.

        Args:
            url: The URL to query to get core configuration information from ARDI
            assetid: The source ID that this core represents"""    
        self.url = url
        self.asset = assetid
    
    def SetARDIDriver(self, driver):
        """Set Driver

        This function associates the given driver with this core

        Args:
            driver: Associate this core with a driver"""       
        self.driver = driver
        self.driver.asset = self.asset
        driver.core = self
                 
    def Close(self):
		"""Closes to connection to the data store"""
		self.driver.Disconnect()
	
    def Reload(self):
        """Force the core to reload

        This function is usually called after a change to the required data"""
        self.LoadConfig()
        
    def LoadConfig(self):
        """Reload driver configuration

        This function triggers the driver to reload it's details from the ARDI server. Note that this
        is very different from the same function located in the DriverBase class. The data returned by
        this query includes all of the individual data points to be sampled from the data source."""
        if (self.source != ''):
			fl = open(self.source,'r')
			xml = fl.read()
			fl.close()
			self.parseconfig(xml)
			return
			
        response = StringIO()
        
        fullurl = "http://" + self.url + "/api/sourcedata?asset=" + str(self.asset) + "&mode=1"
        
        c = pycurl.Curl()
        c.setopt(c.URL,fullurl)
        c.setopt(c.WRITEFUNCTION,response.write)
        c.setopt(c.TIMEOUT,5)
        c.setopt(c.CONNECTTIMEOUT,5)
        c.perform()
        c.close()
        
        self.parseconfig(response.getvalue())
        
    def parseconfig(self,xml):
        """Parse the configuration details

        This function processes the XML from 'LoadConfig' to setup the driver and the many data points."""
        result = xmltodict.parse(xml)
        
        #print "Hello There!"
        
        #Core Parameters
        self.driveraddr = result['sourcedata']['connection']['address']
        #print 'Driver Location: ' + str(self.driveraddr)
        self.hostport = int(result['sourcedata']['connection']['port'])
        self.driver.SetAddress(self.driveraddr)

        if self.driveraddr is None:
            self.driveraddr = ""

        #Data Points
        try:            
            self.pointlist = []
            if '@address' in result['sourcedata']['points']['point']:
                #There is a Single Point
                newpoint = pointinfo()
                point = result['sourcedata']['points']['point']
                newpoint.address = point['@address']
                newpoint.code = point['@code']
                newpoint.transform = point['@transform']
                newpoint.changed = False
                newpoint.value = ""
                newpoint.intvalue = ""
                self.pointlist.append(newpoint)
            else:
                for point in result['sourcedata']['points']['point']:
                    #Add Data Points to List
                    newpoint = pointinfo()
                    newpoint.address = point['@address']
                    newpoint.code = point['@code']
                    newpoint.transform = point['@transform']
                    newpoint.changed = False
                    newpoint.value = ""
                    newpoint.intvalue = ""
                    self.pointlist.append(newpoint)
        except:
            pass
    
    def Query(self, points,start,end,function,grain,first):		
		"""Performs a query on a single source

        \param points: A list of points to query
        \param start: The start date for the query, in YYYY-MM-DD hh:mm:ss format
        \param end: The start date for the query, in YYYY-MM-DD hh:mm:ss format
        \param function: The function to perform. Functions include 'raw', 'avg', 'sum', 'min' and 'max'
        \param grain: The number of seconds between samples, or '0' for natural sampling rate
        \param first: True if this the data is to include the samples immediately preceeding the start time.
		"""
		#print "Starting From " + str(start)
		#print "Ending On " + str(end)

		finalpoints = []
		for pnt in points:
			for itm in self.pointlist:				
				if itm.code == pnt:
					finalpoints.append(itm.address)
					break

		#print "Known Points: "
		#print points
				
		if len(finalpoints) == 0:
			return "ERROR: No valid point specified"
		return self.driver.Query(finalpoints,start,end,function,grain,first)

    def Launch(self,address,asset):
        """Starts the driver

        This function initiates the driver.

        Args:
            address: The encoded destination address (ie. IP address) of the device
            asset: The asset id that this source represents"""
        self.SetARDIDetails(address,asset)
        self.LoadConfig()    
        self.driver.Connect()
                    